/*
	Shizuku
	Kizuato

	for LSI-C 86 ShishokuBan
*/
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>

char	*tmpname1 = "@@TMP1@@.DAT";
char	*tmpname2 = "@@TMP2@@.DAT";

void decodeXOR( FILE *in, unsigned long size, FILE *out, unsigned char x )
{
	unsigned long	i;
	unsigned char	tmp[1];

	for( i = 0; i != size; i++ )
	{
		fread( tmp, 1, 1, in );
		fprintf( out, "%c", tmp[0] ^ x );
	}
	return;
}

unsigned long decodeLZ( FILE *in, unsigned long size, FILE *out )
{
	unsigned char	buf[256];
	unsigned char	*dic;
	unsigned long	dicpos = 0x0FEE;
	unsigned char	byte;
	unsigned char	sign;
	unsigned int	counter;
	unsigned int	decode_startaddr;
	unsigned int	decode_length;
	int				i;
	int				q;
	unsigned long	decode_byte = 0L;
	unsigned long	read_byte = 0L;
	unsigned long	addr = 0L;

	dic = ( unsigned char * )calloc( 1, 4096 );

	while( read_byte < size )
	{
		/* compress sign */
		fread( buf, 1, 1, in ); sign = buf[0];
		read_byte++;

		for( i = 0; i != 8; i++ )
		{
			if( sign & 128 )
			{
				fread( buf, 1, 1, in ); byte = buf[0];
				dic[dicpos] = byte;
				fprintf( out, "%c", byte );
				dicpos++; dicpos &= 0x0FFF;
				read_byte++; decode_byte++;
			} else {
				fread( buf, 2, 1, in ); counter = *( unsigned int *)( buf );
				decode_length = ( counter & 15 ) + 3;		/* 4bit */
				decode_startaddr = ( counter >> 4 );		/* 12bit */
				read_byte += 2;

				/* decode */
				for( q = 0; q != decode_length; q++ )
				{
					addr = decode_startaddr + q; addr &= 0x0FFF;
					dic[dicpos] = dic[addr];
					fprintf( out, "%c", dic[addr] );
					dicpos++; dicpos &= 0x0FFF;
					decode_byte++;
				}
			}
			sign <<= 1;
			if( read_byte > size ) break;
		}
	}
	free( dic );
	return decode_byte;
}

void main( int argc, char **argv )
{
	FILE			*in;
	FILE			*out;
	unsigned char	tmp[256];
	unsigned long	filesize_original = 0L;
	unsigned long	filesize_compress = 0L;

	printf( "Leaf MUSIC Decompress tool 070411 (c)Fac.KUROHANE\n" );

	if( argc == 1 )
	{
		printf( "usage: SZEXT <filename>\n" );
		return;
	}

	in = fopen( argv[1], "rb" );
	if( in == NULL )
	{
		printf( "error: can't open %s\n", argv[1] );
		return;
	}

	printf( "filename: %s\n", argv[1] );

	/* filesize */
	fseek( in, 0, SEEK_END );
	filesize_compress = ftell( in );
	fseek( in, 0, SEEK_SET );

	printf( "filesize_compress: %lu\n", filesize_compress );

	/* header */
	fread( tmp, 6, 1, in );

	if( tmp[3] != 0x00 && tmp[4] != 0x00 )
	{
		printf( "error: not LZ file.\n" );
		goto end;
	}

	filesize_original = *( unsigned long *)( tmp );
	printf( "filesize_original: %lu\n", filesize_original );

	out = fopen( tmpname1, "wb+" );
	if( out == NULL )
	{
		printf( "error: can't write tempfile.\n" );
		goto end;
	}

	/* XOR decode */
	fseek( in, 0x04, SEEK_SET);
	decodeXOR( in, filesize_compress - 4, out, 0xFF );
	fclose( out );
	fclose( in );

	in = fopen( tmpname1, "rb" );
	if( in == NULL )
	{
		printf( "error: can't write tempfile.\n" );
		return;
	}

	out = fopen( tmpname2, "wb+" );
	if( out == NULL )
	{
		printf( "error: can't write tempfile.\n" );
		goto end;
	}

	/* LZ Decode */
	decodeLZ( in, filesize_compress, out );
	fclose( out );

	remove( argv[1] );
	remove( tmpname1 );
	rename( tmpname2, argv[1] );

	printf( "complete.\n" );

end:;
	fclose( in );
	return;
}
